#include "position_weight_matrix.h"
#include "bbq_tokenizer.h"
#include <list>
#include <math.h>


typedef std::list<float> float_list;
//typedef boost::char_separator<char> boost_char_sep;
//typedef boost::tokenizer<boost_char_sep> boost_tokenizer;
//typedef boost::tokenizer<boost_char_sep>::iterator boost_tok_iterator;

float _min(float a, float b, float c, float d)
{
	float min = a;
	if (b<=min)
		min = b;
	if (c<=min)
		min = c;
	if (d<=min)
		min = d;
	return min;
}

float _max(float a, float b, float c, float d)
{
	float max = a;
	if (b>=max)
		max = b;
	if (c>=max)
		max = c;
	if (d>=max)
		max = d;
	return max;
}

namespace bbq {

int position_weight_matrix::match_mode = 2;
	
position_weight_matrix::position_weight_matrix()
{
	match_reverse_complement = false;
	length = 0;
	A = NULL;
	C = NULL;
	G = NULL;
	T = NULL;
	information_content = NULL;
	freq = NULL;
}
		
position_weight_matrix::position_weight_matrix(const position_weight_matrix& M)
{
	//clear();
	match_reverse_complement = M.match_reverse_complement;
	length = M.length;
	A = new float[length];
	C = new float[length];
	G = new float[length];
	T = new float[length];

	int i;
	for (i=0; i<length; i++)
	{
		A[i] = M.A[i];
		C[i] = M.C[i];
		G[i] = M.G[i];
		T[i] = M.T[i];
	}
	
	init();
}

position_weight_matrix& position_weight_matrix::operator=(const position_weight_matrix& M)
{
	clear();
	match_reverse_complement = M.match_reverse_complement;
	length = M.length;
	A = new float[length];
	C = new float[length];
	G = new float[length];
	T = new float[length];
	int i;
	for (i=0; i<length; i++)
	{
		A[i] = M.A[i];
		C[i] = M.C[i];
		G[i] = M.G[i];
		T[i] = M.T[i];
	}

	init();
}

void position_weight_matrix::init()
{
	information_content = new float[length];
	min_freq = new float[length];
	max_freq = new float[length];
	freq = new float*[CHAR_MAX];
	compl_table = new char[256];

	int i;
	for(i=0; i<256; i++) compl_table[i]=(int)'*';
	
	compl_table[(int)'A']='T';
	compl_table[(int)'C']='G';
	compl_table[(int)'G']='C';
	compl_table[(int)'T']='A';

	for (i=0; i<CHAR_MAX; i++)
	{
		freq[i] = NULL;
	}
	reinit();
}

void position_weight_matrix::reinit()
{

	int i;
	for (i=0; i<length; i++)
	{
		min_freq[i] = _min(A[i],C[i],G[i],T[i]);
		max_freq[i] = _max(A[i],C[i],G[i],T[i]);
	}

	freq['A'] = A;
	freq['C'] = C;
	freq['T'] = T;
	freq['G'] = G;
	freq['a'] = A;
	freq['c'] = C;
	freq['t'] = T;
	freq['g'] = G;

	compute_information_content();
}


std::ifstream& operator>>(std::ifstream& ifs, position_weight_matrix& M)
{
	//M.read(ifs);
	return ifs;
}

void position_weight_matrix::clear()
{
	
	if (A!=NULL)
		delete[] A;
	if (C!=NULL)
		delete[] C;
	if (G!=NULL)
		delete[] G;
	if (T!=NULL)
		delete[] T;
	if (freq!=NULL)
		delete[] freq;
	if (information_content!=NULL)
		delete[] information_content;
	A = NULL;
	C = NULL;
	G = NULL;
	T = NULL;
	information_content = NULL;
	freq = NULL;
	
}

void position_weight_matrix::set_reverse_complement(bool rev_comp)
{

	match_reverse_complement = rev_comp;
	// I.: invert order of rows
	/*int l_2 = length / 2;
	int j;
	for (j=0; j<l_2; j++)
	{
		
		float tmp = A[j];
		A[j] = A[length-j-1];
		A[length-j-1] = tmp;
		tmp = C[j];
		C[j] = C[length-j-1];
		C[length-j-1] = tmp;
		tmp = G[j];
		G[j] = G[length-j-1];
		G[length-j-1] = tmp;
		tmp = T[j];
		T[j] = T[length-j-1];
		T[length-j-1] = tmp;
		
	}

	// II.: swap column order
	for (j=0; j<length; j++)
	{
		
		float tmp = A[j];
		A[j] = T[j];
		T[j] = tmp;	
		tmp = C[j];
		C[j] = G[j];
		G[j] = tmp;	
		
	}

	reinit();*/
	
}

bool position_weight_matrix::read(const char* fname)
{
	
	int j=0;
	std::ifstream ifs(fname);

	std::string line;
	std::string filename;
	float_list l_A,l_C,l_G,l_T;
	// read A line

	if (ifs.good())
	{
		getline(ifs,line);
		//std::cout<<line<<"\n";
		tokenizer tokens(line.c_str());
		tokenizer::iterator tok_it = tokens.begin();
		while (tok_it != tokens.end())
		{
			std::string tok_str = *tok_it;
			//std::cout<<tok_str<<"\n";
			float f_val = atof(tok_str.c_str());
			//std::cout<<f_val<<"\n";
			l_A.push_back(f_val);
			tok_it++;
		}
		length = l_A.size();
	} else {
		return false;
	}
	
	float_list::iterator A_it = l_A.begin();
	float_list::iterator A_end = l_A.end();
	A = new float[length];
	for (j=0; j<length; A_it++, j++)
	{
		A[j] = *A_it;
		//std::cout<<A[j]<<", ";
	}

	// read C line
	if (ifs.good())
	{
		getline(ifs,line);
		//std::cout<<line<<"\n";
		tokenizer tokens(line.c_str());
		tokenizer::iterator tok_it = tokens.begin();
		while (tok_it != tokens.end())
		{
			std::string tok_str = *tok_it;
			//std::cout<<tok_str<<"\n";
			float f_val = atof(tok_str.c_str());
			//std::cout<<f_val<<"\n";
			l_C.push_back(f_val);
			tok_it++;
		}
		length = l_C.size();
	} else {
		return false;
	}
	float_list::iterator C_it = l_C.begin();
	float_list::iterator C_end = l_C.end();
	C = new float[length];
	for (j=0; j<length; C_it++, j++)
	{
		C[j] = *C_it;
		//std::cout<<C[j]<<", ";
	}

	// read G line
	if (ifs.good())
	{
		getline(ifs,line);
		//std::cout<<line<<"\n";
		tokenizer tokens(line.c_str());
		tokenizer::iterator tok_it = tokens.begin();
		while (tok_it != tokens.end())
		{
			std::string tok_str = *tok_it;
			//std::cout<<tok_str<<"\n";
			float f_val = atof(tok_str.c_str());
			//std::cout<<f_val<<"\n";
			l_G.push_back(f_val);
			tok_it++;
		}
		length = l_G.size();
	} else {
		return false;
	}
	float_list::iterator G_it = l_G.begin();
	float_list::iterator G_end = l_G.end();
	G = new float[length];
	for (j=0; j<length; G_it++, j++)
	{
		G[j] = *G_it;
		//std::cout<<G[j]<<", ";
	}

	// read T line
	if (ifs.good())
	{
		getline(ifs,line);
		//std::cout<<line<<"\n";
		tokenizer tokens(line.c_str());
		tokenizer::iterator tok_it = tokens.begin();
		while (tok_it != tokens.end())
		{
			std::string tok_str = *tok_it;
			//std::cout<<tok_str<<"\n";
			float f_val = atof(tok_str.c_str());
			//std::cout<<f_val<<"\n";
			l_T.push_back(f_val);
			tok_it++;
		}
		length = l_T.size();
	} else {
		return false;
	}
	float_list::iterator T_it = l_T.begin();
	float_list::iterator T_end = l_T.end();
	T = new float[length];
	for (j=0; j<length; T_it++, j++)
	{
		T[j] = *T_it;
		//std::cout<<T[j]<<", ";
	}

	if (match_mode==0 || match_mode==1)
	{
		for (j=0; j<length; j++)
		{
			
			float sum = A[j]+C[j]+G[j]+T[j];
			if (sum==0.)
				sum = 1.;
			
			A[j] /= sum;
			C[j] /= sum;
			G[j] /= sum;
			T[j] /= sum;
			//std::cout<<"j = "<<j<<"; "<<A[j]<<" "<<C[j]<<" "<<G[j]<<" "<<T[j]<<"\n";
		}
	}

	init();
	
	return true;
}

void position_weight_matrix::compute_information_content()
{
	int i;
	
	for (i=0; i<length; i++)
	{
		float ic_i  = A[i]==0. ? 0. : A[i]*log(4.*A[i]);
		ic_i += C[i]==0. ? 0. : C[i]*log(4.*C[i]);
		ic_i += G[i]==0. ? 0. : G[i]*log(4.*G[i]);
		ic_i += T[i]==0. ? 0. : T[i]*log(4.*T[i]);
		information_content[i] = ic_i;		
	}
}

void position_weight_matrix::set_matches(const char* seq, float threshold)
{
	if (match_mode==0 || match_mode==1)
		set_naive_matches(seq, threshold);
	else if (match_mode==2)
		set_MATCH_matches(seq, threshold);
}

void position_weight_matrix::set_MATCH_matches(const char* seq, float threshold)
{
	//std::cout<<"moin.\n"; std::cout.flush();
	matches.clear();
	max_score = 0.;
	max_pos = 0;
	bool curr_rev;
	max_score_rev = false;
	
	int slen = strlen(seq);
	int max_idx = slen-length+1;
	int i,j;
	for (i=0; i<max_idx; i++)
	{
		float score,rev_score,Min,Max,Current,rev_Current;
		score = 0.;
		rev_score = 0.;
		Min = 0.;
		Max = 0.;
		Current = 0.;
		rev_Current = 0.;
		for (j=0; j<length; j++)
		{
			Min += information_content[j]*min_freq[j];
			Max += information_content[j]*max_freq[j];
			if (freq[seq[i+j]]!=NULL)
			{
				Current += information_content[j]*freq[seq[i+j]][j];
				rev_Current += information_content[length-j-1]*freq[compl_table[seq[i+j]]][length-j-1];
			}
		}
		score = (Current-Min) / (Max-Min);
		rev_score = (rev_Current-Min) / (Max-Min);
		if (score>=threshold)
			// biologists' sequences start with one, so we add 1 to our position...
			matches.push_back(match(i+1,score,false));
		if (rev_score>=threshold && match_reverse_complement)
			matches.push_back(match(i+1,rev_score,true));
		//std::cout<<"fwd score "<<score<<" vs max score "<<max_score<<"\n";
		if (score>max_score)
		{
			max_score_rev = false;
			max_score = score;
			max_pos = i;
		}
		//std::cout<<"rev score "<<score<<" vs max score "<<max_score<<"\n";
		if (rev_score>max_score && match_reverse_complement)
		{
			max_score_rev = true;
			max_score = rev_score;
			max_pos = i;
		}
	}
}


void position_weight_matrix::set_naive_matches(const char* seq, float threshold)
{
	//std::cout<<"moin.\n"; std::cout.flush();
	matches.clear();
	float max_score = 0.;
	int max_pos = 0;
	
	int slen = strlen(seq);
	int max_idx = slen-length;
	int i,j;
	for (i=0; i<max_idx; i++)
	{
		float score;
		if (match_mode==0)
			score = 0.;
		else
			score = 1.;

		for (j=0; j<length; j++)
		{
			float new_score;
			switch (seq[i+j])
			{
				case 'a':
				case 'A': 
					new_score = A[j];
					break;
				case 'c':
				case 'C':
					new_score = C[j];
					break;
				case 'g':
				case 'G':
					new_score = G[j];
					break;
				case 't':
				case 'T':
					new_score = T[j];
					break;
				default:
					new_score = 0.;
					break;
			}
			if (match_mode==0)
				score += new_score;
			else
				score *= new_score;				
		}
		if (score>=threshold)
			matches.push_back(match(i,score,false));
		if (score>=max_score)
		{
			max_score = score;
			max_pos = i;
		}
	}
}



position_weight_matrix::match_iterator position_weight_matrix::begin()
{
	return matches.begin();
}

position_weight_matrix::match_iterator position_weight_matrix::end()
{
	return matches.end();
}


position_weight_matrix::match_list position_weight_matrix::get_match_list()
{
	return matches;
}

std::ostream& operator<<(std::ostream& ofs, const position_weight_matrix& M)
{
	int l = M.length;
	int j;
	for (j=0; j<l; j++)
		ofs<<M.A[j];
	ofs<<std::endl;
	for (j=0; j<l; j++)
		ofs<<M.C[j];
	ofs<<std::endl;
	for (j=0; j<l; j++)
		ofs<<M.G[j];
	ofs<<std::endl;
	for (j=0; j<l; j++)
		ofs<<M.T[j];
	ofs<<std::endl;
	return ofs;
}

}

