#ifndef POSITION_WEIGHT_MATRIX_H
#define POSITION_WEIGHT_MATRIX_H


#include <fstream>
#include <list>
#include <iostream>
#include "triple.h"

namespace bbq {

class position_weight_matrix
	{
		
	public:

		/// first indicates the position, second determines the match
		/// score, while third indicates the direction of the match
		/// (false iff reverse match)
		typedef std::triple<int,float,bool> match;
		
		typedef std::list< match > match_list;
		
		typedef match_list::iterator match_iterator;
		
		static int match_mode;

		/// true iff both forward matches and reverse complement matches
		/// are to be computed
		bool match_reverse_complement;
		
		float* A;
		float* C;
		float* G;
		float* T;

		float* min_freq;
		float* max_freq;
		float* information_content;

		char* compl_table;
		
		float** freq;

		int length;

		float max_score;
		float max_score_rev;
		
		int max_pos;

		position_weight_matrix();
		
		position_weight_matrix(const position_weight_matrix&);

		position_weight_matrix& operator=(const position_weight_matrix&);

		friend std::ifstream& operator>>(std::ifstream& ifs, position_weight_matrix& M);

		match_list matches;
		
		bool read(const char* fname);

		void set_matches(const char* seq, float threshold);

		void set_MATCH_matches(const char* seq, float threshold);

		void set_naive_matches(const char* seq, float threshold);
		
		void set_reverse_complement(bool);
		
		void compute_information_content();

		void clear();

		void init();

		void reinit();

		match_list get_match_list();

		match_iterator begin();
		match_iterator end();
		
	};


std::ifstream& operator>>(std::ifstream& ifs, position_weight_matrix& M);

std::ostream& operator<<(std::ostream& ofs, const position_weight_matrix& M);

std::ostream& operator<<(std::ostream& ofs, const position_weight_matrix::match& M);


}


#endif

