#include <ctype.h>
#include <iostream>
#include <fstream>
#include <list>
#include <stdio.h>
#include <string>
#include <utility>
#include <map>
#include <set>
//#include <boost/filesystem/operations.hpp>
//#include <boost/filesystem/fstream.hpp> 
//#include <boost/tokenizer.hpp>
//#include <boost/lexical_cast.hpp>

#include "position_weight_matrix.h"

// libsequence headers for reading fasta files
#include "fasta.h"

typedef bbq::position_weight_matrix PWM;
typedef std::list<std::string> str_list;
typedef std::list<PWM> PWM_list;
typedef std::list<float> float_list;
typedef std::list<int> int_list;
typedef std::list< std::pair<float,std::string> > float_str_list;
typedef PWM::match_list match_list;
typedef PWM::match match;
typedef std::map<std::string, int> str_to_int_map;
typedef std::set< std::pair<float,std::string> > float_str_set;

PWM M;
PWM M_rev;
float threshold;

int mode;

str_list seq_list;
str_list name_list;

bool reverse_complement;
bool quiet;
bool 	U_equals_T;
int num_of_opts;
int num_of_matrices;
const char* fasta_filename;
const char* matrix_filename;

float_str_list matrix_score_name_list;
PWM_list matrix_list;
str_to_int_map best_match_pos;

bool list_mode;

int global_argc;
char** global_argv;

bool read_matrix(const char* fname, PWM& M)
{
	
	M.read(fname);
	
}

void substitute_UT(std::string& str)
{
	int i =0;
	int len = str.size();
	for (;i<len; i++)
		if (str[i]=='U')
			str[i] = 'T';
}

bool read_sequences(const char* fname)
{

	fasta_reader fastafile (fname);    
	std::string header,seq;

	while (fastafile.get_seq (header, seq)) {
		if (U_equals_T)
			substitute_UT(seq);
		std::string T_name = header;
		name_list.push_back(header);
		seq_list.push_back(seq);
	} 

}

void perform_standard_matches()
{
	read_matrix(matrix_filename, M);
	
	if (reverse_complement)
		M.set_reverse_complement(true);
	
	//M_rev = M;
	//M_rev.reverse_complement();
	//std::cerr<<M<<std::endl<<M_rev<<std::endl;

	read_sequences(fasta_filename);

	str_list::iterator s_it = seq_list.begin();
	str_list::iterator s_end = seq_list.end();
	str_list::iterator n_it = name_list.begin();
	str_list::iterator n_end = name_list.end();

	int j;
	for (; s_it!=s_end; s_it++, n_it++)
	{
		if (!quiet)
			std::cout<<"# matches in sequence "<<*n_it<<":\n";
		//else
		//	std::cout<<"# "<<*n_it<<":\n";
		M.set_matches(s_it->c_str(), threshold);
		match_list ml = M.get_match_list();
		match_list::iterator m_it = ml.begin();
		match_list::iterator m_end = ml.end();
		if (!quiet)
		{	
			std::cout<<"# "<<ml.size()<<" matches found.\n";
			std::cout<<"# max score = "<<M.max_score<<" at pos. "<<M.max_pos<<"; matching seq = ";
			for (j=0; j<M.length; j++)
				std::cout<<(*s_it)[M.max_pos+j];
			std::cout<<"\n";
		}			
		for (; m_it!=m_end; m_it++)
		{
			if (!m_it->third)
				std::cout<<m_it->first<<" "<<m_it->second<<" "<<global_argv[num_of_opts+2]<<" fwd "<<*n_it<<"\n";
			else
				std::cout<<m_it->first<<" "<<m_it->second<<" "<<global_argv[num_of_opts+2]<<" rev "<<*n_it<<"\n";				
			if (!quiet)
			{
				std::cout<<"# pos "<<m_it->first<<" weight "<<m_it->second<<"; matching seq = ";
				for (j=0; j<M.length; j++)
					std::cout<<(*s_it)[m_it->first+j];
				std::cout<<"\n";
			}
		}

		/*if (reverse_complement)
		{
			if (!quiet)
				std::cout<<"# reverse complement matches in sequence "<<*n_it<<":\n";
			//else
			//	std::cout<<"# rev "<<*n_it<<":\n";
			M_rev.set_matches(s_it->c_str(), threshold);
			match_list ml = M_rev.get_match_list();
			m_it = ml.begin();
			m_end = ml.end();
			if (!quiet)
			{
				std::cout<<"# "<<ml.size()<<" reverse complement matches found.\n";
				std::cout<<"# max reverse complement score = "<<M_rev.max_score<<" at pos. "<<M_rev.max_pos<<"; matching seq = ";
				for (j=0; j<M_rev.length; j++)
					std::cout<<(*s_it)[M_rev.max_pos+j];
				std::cout<<"\n";
			}
			for (; m_it!=m_end; m_it++)
			{
				std::cout<<m_it->first<<" "<<m_it->second<<" "<<argv[num_of_opts+2]<<" rev "<<*n_it<<"\n";
				if (!quiet)
				{
					std::cout<<"# * pos "<<m_it->first<<" weight "<<m_it->second<<"; matching seq = ";
					for (j=0; j<M_rev.length; j++)
						std::cout<<(*s_it)[m_it->first+j];
					std::cout<<"\n";
				}
			}
		}*/
	}

}

void perform_list_matches()
{
	
	if (fasta_filename==NULL) {
		std::cerr<<"ERROR: fasta file needs to be specified with option -F <filename> when using list mode -L";
		exit(-1);
	} 
	read_sequences(fasta_filename);

	PWM_list::iterator M_it = matrix_list.begin();
	PWM_list::iterator M_end = matrix_list.end();
	float_str_list::iterator N_it = matrix_score_name_list.begin();
	float_str_list::iterator N_end = matrix_score_name_list.end();

	for (; M_it!=M_end; M_it++, N_it++)
	{
		//std::cerr<<"reading "<<N_it->second<<"\n";
		read_matrix(N_it->second.c_str(),*M_it);
		
		if (reverse_complement)
			M_it->set_reverse_complement(true);
		else
			M_it->set_reverse_complement(false);
		
		str_list::iterator s_it = seq_list.begin();
		str_list::iterator s_end = seq_list.end();
		str_list::iterator n_it = name_list.begin();
		str_list::iterator n_end = name_list.end();
		
		int j;
		N_it->first = 0.;
		best_match_pos[N_it->second] = -1;
		
		for (; s_it!=s_end; s_it++, n_it++)
		{
			M_it->set_matches(s_it->c_str(), 1.);
			if (M_it->max_score > N_it->first)
			{
				N_it->first = M_it->max_score;
				best_match_pos[N_it->second] = M_it->max_pos;
			}
		}

	}

	float_str_set result_set;
	N_it = matrix_score_name_list.begin();
	N_end = matrix_score_name_list.end();
	for (; N_it!=N_end; N_it++)
		result_set.insert(*N_it);
	
	float_str_set::iterator R_it = result_set.begin();
	float_str_set::iterator R_end = result_set.end();
	for (; R_it != R_end; ++R_it)
	{
		std::cout<<R_it->second<<": "<<R_it->first<<"\n";
	}


}

int main(int argc, char** argv)
{
	global_argc = argc;
	global_argv = argv;
	
	if (argc<4)
	{
		std::cout<<"usage 1: pwmatch [options] <threshold> <matrixfile> <fastafile>\n";
		std::cout<<"usage 2: pwmatch [options] -L <matrixfile1> ... <matrixfileN> -F <fastafile>\n";		
		std::cout<<"\noptions:\n";
		std::cout<<"  -M  : Use Transf*** MATCH scores (default)\n";
		std::cout<<"  -s  : sum up column scores\n";
		std::cout<<"  -m  : multiply column scores\n";
		std::cout<<"  -r  : match matrix as well as its reverse complement\n";
		std::cout<<"  -q  : quiet mode (avoid verbous output)\n";
		std::cout<<"  -TU : treat T not equal to U (default: T=U)\n";
		return 1;
	}

	reverse_complement = false;
	quiet = false;
	num_of_opts = 0;
	num_of_matrices = 0;
	fasta_filename = NULL;
	list_mode = false;
	U_equals_T = true;
	
	PWM::match_mode = 2;
	int i;
	for (i=1; i<argc && argv[i][0]=='-'; i++)
	{
		if (argv[1][0]=='-')
		{
			num_of_opts++;
			switch(argv[i][1])
			{
				case 'q':
					quiet = true;
					break;
				case 'r':
					reverse_complement = true;
					break;
				case 's':
					PWM::match_mode = 0;
					break;
				case 'm':
					PWM::match_mode = 1;
					break;
				case 'M':
					PWM::match_mode = 2;
					break;
				case 'L':
					// read matrix file list
					list_mode = true;
					num_of_matrices = 0;
					i++;
					for (; i<argc && argv[i][0]!='-'; i++, num_of_opts++) {
						matrix_score_name_list.push_back(std::pair<float,std::string>(0.,std::string(global_argv[i])));
						matrix_list.push_back(PWM());
					}
					i--;
					num_of_opts--;
					break;
				case 'F':
					// read fasta filename
					list_mode = true;
					fasta_filename = argv[i+1];
					i++;
					num_of_opts++;
					break;
				case 'T':
					// -TU makes U==T
					U_equals_T = false;
					break;
				case '-':
					if (!strcmp(argv[i],"--aT=kaU"))
					{
						std::cerr<<"Des wuesst i a wenn a T a U wer.\n";
						U_equals_T = false;
					}
				default:
					std::cout<<"unknown option: -"<<argv[i][1]<<"\n";
					break;
			}
		}
	}


	if (!list_mode) {
		threshold = atof(argv[num_of_opts+1]);
		fasta_filename = argv[num_of_opts+3];
		matrix_filename = argv[num_of_opts+2];
		perform_standard_matches();
	} else {
		perform_list_matches();
	}
	
}

